@@ -16,6 +16,7 @@
 
 #include "system.h"
 #include <setjmp.h>
+#include <errno.h>
 
 /* locally defined functions */
 static void split(ptr k, ptr *s);
@@ -27,6 +28,12 @@ static void keyboard_interrupt(ptr tc);
 
 static void (*register_modified_signal)(int);
 
+#ifdef WIN32
+typedef int *(*get_errno_ptr_t)(void);
+static get_errno_ptr_t msvcrt_get_errno_ptr;
+static get_errno_ptr_t ucrt_get_errno_ptr;
+#endif
+
 ptr S_get_scheme_arg(ptr tc, iptr n) {
 
     if (n <= asm_arg_reg_cnt) return REGARG(tc, n);
@@ -464,6 +471,63 @@ void S_handle_mvlet_error(void) {
     handle_call_error(tc, ERROR_MVLET, Sfalse);
 }
 
+#ifdef WIN32
+static int *get_errno_ptr(void) {
+  return &errno;
+}
+#endif
+
+ptr S_save_errno(void) {
+    int errno_val;
+
+#ifdef WIN32
+    {
+      ptr tc = get_thread_context();
+      if (CURRENTERRNOSOURCE(tc) == Sfalse) {
+	errno_val = errno;
+      } else if (Schar_value(STRIT(SYMNAME(CURRENTERRNOSOURCE(tc)), 0)) == 'm' /* msvcrt */) {
+	if (!msvcrt_get_errno_ptr) {
+	  HMODULE hm;
+	  get_errno_ptr_t new_get_errno_ptr;
+	  hm = LoadLibrary("msvcrt.dll");
+	  if (hm)
+	    new_get_errno_ptr = (get_errno_ptr_t)GetProcAddress(hm, "_errno");
+	  if (!new_get_errno_ptr)
+	    new_get_errno_ptr = get_errno_ptr;
+	  while (msvcrt_get_errno_ptr == NULL) {
+	    COMPARE_AND_SWAP_PTR(&msvcrt_get_errno_ptr, NULL, new_get_errno_ptr);
+	  }
+	}
+	errno_val = *(msvcrt_get_errno_ptr());
+      } else {
+	if (!ucrt_get_errno_ptr) {
+	  HMODULE hm;
+	  get_errno_ptr_t new_get_errno_ptr;
+	  hm = LoadLibrary("ucrtbase.dll");
+	  if (hm)
+	    new_get_errno_ptr = (get_errno_ptr_t)GetProcAddress(hm, "_errno");
+	  if (!new_get_errno_ptr)
+	    new_get_errno_ptr = get_errno_ptr;
+	  while (ucrt_get_errno_ptr == NULL) {
+	    COMPARE_AND_SWAP_PTR(&ucrt_get_errno_ptr, NULL, new_get_errno_ptr);
+	  }
+	}
+	errno_val = *(ucrt_get_errno_ptr());
+      }
+    }
+#else
+    errno_val = errno;
+#endif
+
+    return Sinteger(errno_val);
+}
+
+#ifdef WIN32
+ptr S_save_last_error(void) {
+    return Sinteger(GetLastError());
+}
+#endif
+
 void S_handle_event_detour() {
     ptr tc = get_thread_context();
     ptr resume_proc = CP(tc);
